__version__ = "2.0.1"
# __version__ has to be defined in the first line

from .classifiers import (Box_Plot, Equal_Interval, Fisher_Jenks,
                          Fisher_Jenks_Sampled, HeadTail_Breaks, Jenks_Caspall,
                          Jenks_Caspall_Forced, Jenks_Caspall_Sampled,
                          Max_P_Classifier, Maximum_Breaks, Natural_Breaks,
                          Quantiles, Percentiles, Std_Mean, User_Defined,
                          load_example, gadf, K_classifiers, CLASSIFIERS)

